"""Classes for holding and processing Heatmier heating and hot water schedule fields"""
import logging
import itertools

from hm_constants import CURRENT_TIME_DAY, CURRENT_TIME_HOUR, CURRENT_TIME_MIN

#mapping for chunks of heating schedule for a day
MAP_HOUR = 0
MAP_MIN = 1
MAP_TEMP = 2
#hour that indicates unused item
HOUR_UNUSED = 24
#index for returned schedule
SCH_ENT_DAY = 0
SCH_ENT_HOUR = 1
SCH_ENT_MIN = 2
SCH_ENT_TEMP = 3
#useful constants
HOUR_MINUTES = 60

class Scheduler(object):
    """General Schedule base class, providing a set of inherited methods"""
    #entry is a day or week/end, item is a part within the entry
    fieldbase = None
    fieldnames = ['mon', 'tues', 'wed', 'thurs', 'fri', 'sat', 'sun', 'wday', 'wend']

    def __init__(self):
        if self.fieldbase is not None:
            self.entrynames = [x + self.fieldbase for x in self.entrynames]
            self.fieldnames = [x + self.fieldbase for x in self.fieldnames]
        self.entries = dict.fromkeys(self.fieldnames, None)

    def set_raw_all(self, schedule):
        """Set all fields to same schedule"""
        for entry in self.fieldnames:
            self.set_raw(entry, schedule)

    def set_raw(self, entry, schedule):
        """Set single field to schedule"""
        if not entry in self.fieldnames:
            raise ValueError('Schedule entry does not exist %s'%entry)
        if not len(schedule) is self.valuesperentry * self.entriesperday:
            raise ValueError('Schedule entry wrong length %i'%len(entry))
        self.entries[entry] = schedule

    def set_raw_field(self, field):
        """Set single field to schedule from field pointer"""
        self.set_raw(field.name, field.value)
        
    def get_entry_names(self, entryname):
        """Get list of field names. If single returns full field name from short.
        Expects short name input or 'all'."""
        if entryname == 'all':
            return self.entrynames
        else:
            if self.fieldbase is not None:
                entryname = entryname + self.fieldbase
            if not entryname in self.entrynames:
                raise ValueError('Schedule entry not setable or does not exist %s'%entryname)
            return [entryname]
        
    def pad_schedule(self, schedule):
        """Pads a partial schedule up to correct length"""
        if not len(schedule)%self.valuesperentry == 0:
            raise IndexError("Schedule length not multiple of %d"%self.valuesperentry)
        pad_item = [HOUR_UNUSED, 0, 12][0:self.valuesperentry]

        return schedule + pad_item * ((self.valuesperentry * self.entriesperday - len(schedule))/self.valuesperentry)

    def display(self):
        """Prints schedule to stdout"""
        print(self.title + " Schedule")

        for name, entry in itertools.izip(self.printnames, self.entrynames):
            if self.entries[entry] is None:
                textstr = "None"
            else:
                textstr = self.entry_text(self.entries[entry])
            print(name.ljust(10) + textstr)
            logging.info(textstr)

    @staticmethod
    def _chunks(fulllist, chunklength):
        """Yield successive n-sized chunks from l."""
        for pos in range(0, len(fulllist), chunklength):
            yield fulllist[pos:pos + chunklength]
    
    @staticmethod
    def _reversechunks(fulllist, chunklength):
        """Yield successive n-sized chunks from l."""
        for pos in range(len(fulllist)-chunklength, -1, -chunklength):
            yield fulllist[pos:pos + chunklength]
            
    def get_current_schedule_item(self, timearray):
        """Gets the current item from schedule"""
        ####check time and vars current
        
        todayschedule = self._get_schedule_entry(timearray[CURRENT_TIME_DAY])
            
        scheduletarget = self._get_current_item_from_an_entry(todayschedule, timearray)
            
        if scheduletarget is None:
            yestschedule = self._get_previous_schedule_entry(timearray)
            scheduletarget = self._get_last_item_from_an_entry(yestschedule)
            return [self._get_previous_day(timearray)] + scheduletarget
        else:
            return [timearray[CURRENT_TIME_DAY]] + scheduletarget
    
    def get_next_schedule_item(self, timearray):
        """Gets the next item from schedule"""
        todayschedule = self._get_schedule_entry(timearray[CURRENT_TIME_DAY])
        
        scheduletarget = self._get_next_item_from_an_entry(todayschedule, timearray)
            
        if scheduletarget is None:
            tomschedule = self._get_next_schedule_entry(timearray)
            scheduletarget = self._get_first_item_from_an_entry(tomschedule)
            return [self._get_next_day(timearray)] + scheduletarget
        else:
            return [timearray[CURRENT_TIME_DAY]] + scheduletarget
            
    def _get_next_item_from_an_entry(self, schedule, timearray):
        """Gets the next item from a days schedule"""
        scheduletarget = None
        dayminutes = timearray[CURRENT_TIME_HOUR] * HOUR_MINUTES + timearray[CURRENT_TIME_MIN]

        for i in self._reversechunks(schedule, self.valuesperentry):
            if dayminutes < i[MAP_HOUR] * HOUR_MINUTES + i[MAP_MIN] and i[MAP_HOUR] != HOUR_UNUSED:
                scheduletarget = i
        
        return scheduletarget
    
    def _get_current_item_from_an_entry(self, schedule, timearray):
        """Gets the current item from a days schedule"""
        scheduletarget = None
        dayminutes = timearray[CURRENT_TIME_HOUR] * HOUR_MINUTES + timearray[CURRENT_TIME_MIN]
        
        for i in self._chunks(schedule, self.valuesperentry):
            if dayminutes >= i[MAP_HOUR] * HOUR_MINUTES + i[MAP_MIN] and i[MAP_HOUR] != HOUR_UNUSED:
                scheduletarget = i
        
        return scheduletarget
        
    def _get_previous_schedule_entry(self, timearray):
        """Get previous days schedule"""
        return self._get_schedule_entry(self._get_previous_day(timearray))

    def _get_next_schedule_entry(self, timearray):
        """Get next days schedule"""
        return self._get_schedule_entry(self._get_next_day(timearray))

    @staticmethod
    def _get_previous_day(timearray):
        """Get day number for previous day"""
        #shift from 1-7 to 0-6, subtract 1, modulo, shift back to 1-7
        return ((timearray[CURRENT_TIME_DAY] - 1 - 1) % 7) + 1
        
    @staticmethod
    def _get_next_day(timearray):
        """Get next days schedule"""
        #shift from 1-7 to 0-6, add 1, modulo, shift back to 1-7
        return ((timearray[CURRENT_TIME_DAY] - 1 + 1) % 7) + 1
        
    def _get_first_item_from_an_entry(self, schedule):
        """Gets the first item from a days schedule"""
        #gets first schedule entry if valid (not 24)
        firstentry = self._chunks(schedule, self.valuesperentry).next()
        if firstentry[MAP_HOUR] != HOUR_UNUSED:
            return firstentry
        else:
            return None

    def _get_last_item_from_an_entry(self, schedule):
        """Gets the last item from a days schedule"""
        #gets last valid schedule entry (not 24)
        scheduletarget = None
        for i in self._reversechunks(schedule, self.valuesperentry):
            if i[MAP_HOUR] != HOUR_UNUSED:
                scheduletarget = i
                break
        return scheduletarget

class SchedulerDay(Scheduler):
    """Inherited class with day variables and methods"""
    entrynames = ['mon', 'tues', 'wed', 'thurs', 'fri', 'sat', 'sun']
    printnames = ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday']
    
    def _get_schedule_entry(self, day):
        """Determines the schedule for a day"""
        return self.entries[self.entrynames[day - 1]]
    
    def get_entry_names(self, entryname):
        """Get list of field names. If single returns full field name from short.
        Expects short name input or 'all' or 'wday' or 'wend'."""
        if entryname == 'wday':
            return self.entrynames[0:5]
        elif entryname == 'wend':
            return self.entrynames[5:6]
        else:
            return super(SchedulerDay, self).get_entry_names(entryname)

class SchedulerWeek(Scheduler):
    """Inherited class with week variables and methods"""
    entrynames = ['wday', 'wend']
    printnames = ['Weekdays', 'Weekends']
    
    def _get_schedule_entry(self, day):
        """Determines the schedule for a day based on whether weekday or weekend"""
        if day == 6 or day == 7:
            return self.entries[self.entrynames[1]]
        elif day >= 1 and day <= 5:
            return self.entries[self.entrynames[0]]
        else:
            raise ValueError("Day not recognised")
            
class SchedulerHeat(Scheduler):
    """Inherited class with heating variables and methods"""
    title = 'Heating'
    valuesperentry = 3
    entriesperday = 4
    fieldbase = '_heat'
    
    def entry_text(self, data):
        """Assembles string describing a heat schdule entry"""
        tempstr = ''
        for valueset in self._chunks(data, self.valuesperentry):
            if valueset[MAP_HOUR] != HOUR_UNUSED:
                tempstr += "%02d:%02d at %02iC " % (valueset[MAP_HOUR], valueset[MAP_MIN], valueset[MAP_TEMP])
                
        return tempstr
    
class SchedulerWater(Scheduler):
    """Inherited class with hot water variables and methods"""
    title = 'Hot Water'
    valuesperentry = 2
    entriesperday = 8
    fieldbase = '_water'
    
    entry_formats = {
        0: lambda dataset: "On at %02d:%02d " %(dataset[0], dataset[1]), #used for on entries
        1: lambda dataset: "Off at %02d:%02d, " %(dataset[0], dataset[1]) #used for off entries
    }
    group_formats = {
        0: lambda count: "Time %i " %(count), #Group number for time entries
        1: lambda _: ""
    }
    
    def entry_text(self, data):
        """Assembles string describing a water schdule entry"""
        tempstr = ''
        for entry, dataset in enumerate(self._chunks(data, 2)):
            count = int(entry/2) + 1 #group number from entries
            tempstr += self.group_formats[entry % 2](count)
            if dataset[0] != HOUR_UNUSED:
                tempstr += self.entry_formats[entry % 2](dataset)
                
        return tempstr
    
class SchedulerDayHeat(SchedulerDay, SchedulerHeat):
    """Class for day based heating schedule"""
    pass
class SchedulerWeekHeat(SchedulerWeek, SchedulerHeat):
    """Class for week based heating schedule"""
    pass
class SchedulerDayWater(SchedulerDay, SchedulerWater):
    """Class for day based hot water schedule"""
    pass
class SchedulerWeekWater(SchedulerWeek, SchedulerWater):
    """Class for wee based hot water schedule"""
    pass



#!/usr/bin/env python
"""Gets data from Heatmiser stats repeatedly and displays
Ian Horsley 2018
"""

import time
import logging

from heatmisercontroller.logging_setup import initialize_logger_full
from heatmisercontroller.network import HeatmiserNetwork
from heatmisercontroller.exceptions import HeatmiserError

#start logging
initialize_logger_full('logs', logging.INFO)

HMN = HeatmiserNetwork()

def readanddisplay():
    """Read all data from all controllers and print current time"""
    print("\r\nGetting time for %2d in %s *****************************" %(current_controller.set_address, current_controller.set_long_name))
    try:
        current_time = current_controller.read_time()
    except HeatmiserError as err:
        print("C%d in %s Failed to Read due to %s" %(current_controller.set_address, current_controller.name.ljust(4), str(err)))
    else:
        print("C%d in %s time is %s" %(current_controller.set_address, current_controller.name.ljust(4), str(current_time)))

# read and display, without any corrections
for current_controller in HMN.controllers:
    current_controller.set_autocorrectime = False
    readanddisplay()

print("waiting for 30 seconds before fixing any errors")
time.sleep(30) # sleep before next cycle, which will fix any errors

# read and display, and try to correct if wrong
for current_controller in HMN.controllers:
    current_controller.set_autocorrectime = True
    readanddisplay()
        


"""Class to initialise loggers to screen and files"""

import logging
import os

from logging.handlers import RotatingFileHandler

def _add_streamhandler(screenlevel, logger):
    """create console handler and set level to requested level"""
    screenhandler = logging.StreamHandler()
    screenhandler.setLevel(screenlevel)
    screenformatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    screenhandler.setFormatter(screenformatter)
    logger.addHandler(screenhandler)
    logging.debug('Added stream handler')

def _update_streamhandler(screenlevel, screenhandler):
    screenhandler.setLevel(screenlevel)
    screenformatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    screenhandler.setFormatter(screenformatter)
    
def _add_filehandler(output_dir, logger):
    """create error file handler and set level to warn"""
    filehandler = logging.FileHandler(os.path.join(output_dir, "error.log"), "w", encoding=None, delay="true")
    filehandler.setLevel(logging.WARN)
    fileformatter = logging.Formatter("%(asctime)-15s %(levelname)s - %(message)s")
    filehandler.setFormatter(fileformatter)
    logger.addHandler(filehandler)
    logging.debug('Added file handler')

def _add_allfilehandler(output_dir, logger):
    """create debug file handler and set level to debug"""
    #a is append, w is write
    # this one ok is to rehandle because it is an append. Also can be added by a second call to initialize logger
    
    allhandler = RotatingFileHandler(os.path.join(output_dir, "all.log"), mode='a', maxBytes=5*1024*1024,
                                 backupCount=2, encoding=None, delay=0)
    allhandler.setLevel(logging.DEBUG)
    allformatter = logging.Formatter("%(asctime)-15s %(levelname)s - %(message)s")
    allhandler.setFormatter(allformatter)
    logger.addHandler(allhandler)
    logging.debug('Added all handler')
    
def initialize_logger(output_dir, screenlevel):
    """Class to initialise loggers to screen and files
    Returns logger if new loggers added"""
    
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    
    if not logger.handlers: #only create if handlers haven't already been created.
        _add_streamhandler(screenlevel, logger)
        _add_filehandler(output_dir, logger)
    else:
        #check for stream and update if needed
        for handler in logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                _update_streamhandler(screenlevel, handler)
                break
        else:
            _add_streamhandler(screenlevel, logger)
        #check for filehandler
        if not any(isinstance(handler, logging.FileHandler) for handler in logger.handlers):
            _add_filehandler(output_dir, logger)

    return logger
        
def initialize_logger_full(output_dir, screenlevel):
    """Class to initialise loggers to screen and files"""
    
    logger = initialize_logger(output_dir, screenlevel)

    #check for rotating filehandler
    if not any(isinstance(handler, logging.handlers.RotatingFileHandler) for handler in logger.handlers):
        _add_allfilehandler(output_dir, logger)

def csvlist(listitems):
    """function to combine list items into csv string"""
    return ', '.join(map(str, listitems))



"""Handles Heatmiser network of devices and adaptor

also loads adaptor and keeps track of Heatmiser devices

Ian Horsley 2018
"""

import os
import logging

# Import our own stuff
from genericdevice import DEVICETYPES
from generaldevices import HeatmiserBroadcastDevice, ThermoStatUnknown
from adaptor import HeatmiserAdaptor
from hm_constants import SLAVE_ADDR_MIN, SLAVE_ADDR_MAX
from .exceptions import HeatmiserResponseError
import setup as hms

class HeatmiserNetwork(object):
    """Class that connects a set of devices (from configuration) and an adpator."""
    ### stat list setup

    def __init__(self, configfile=None):
        
        # Select default configuration file if none provided
        if configfile is None:
            self._module_path = os.path.abspath(os.path.dirname(__file__))
            configfile = os.path.join(self._module_path, "hmcontroller.conf")
          
        # Initialize controller setup
        try:
            self._setup = hms.HeatmiserControllerFileSetup(configfile)
            settings = self._setup.settings
        except hms.HeatmiserControllerSetupInitError as err:
            logging.error(err)
            raise
        
        # Initialize and connect to heatmiser network, probably through serial port
        self.adaptor = HeatmiserAdaptor(self._setup)
        
        # Load device list from settings or find devices if none listed
        self.controllers = []
        self._addresses_in_use = []
        if 'devices' in settings:
            if len(settings['devices']):
                self._set_stat_list(settings['devices'], settings['devicesgeneral'])
        else: #if devices not defined then auto run find devices.
            self.find_devices()
        
        # Create a broadcast device
        setattr(self, "All", HeatmiserBroadcastDevice(self.adaptor, "Broadcast to All", self.controllers))
        self._current = self.All
      
    def _set_stat_list(self, statlist, generalsettings):
        """Store list of devives and create objects for each"""
        self._statlist = statlist
        self._statnum = len(self._statlist)

        self.controllers = range(self._statnum)
        for name, controllersettings in statlist.iteritems():
            if hasattr(self, name):
                logging.warn("error duplicate stat name")
            else:
                new_device = self.add_device(name, controllersettings, generalsettings)
                self.controllers[controllersettings['display_order'] - 1] = new_device
        self._current = self.controllers[0]
    
    def add_device(self, name, controllersettings, generalsettings=None):
        """Add device to network"""
        expected_model = controllersettings['expected_model']
        expected_prog_mode = controllersettings['expected_prog_mode']
        new_device = DEVICETYPES[expected_model][expected_prog_mode](self.adaptor, controllersettings, generalsettings)
        setattr(self, name, new_device)
        setattr(new_device, 'name', name) #make name avaliable when accessing by id
        self._addresses_in_use.append(controllersettings['address'])
        return new_device
    
    def find_devices(self, max_address=SLAVE_ADDR_MAX):
        """Find devices on the network not in the configuration file"""
        unused_addresses = [address for address in range(SLAVE_ADDR_MIN, max_address + 1) if address not in self._addresses_in_use]
        for address in unused_addresses:
            try:
                controllersettings = {'address': address}
                test_device = ThermoStatUnknown(self.adaptor, controllersettings, self._setup.settings['devicesgeneral'])
                # use fields from device rather to set the expected mode and type
                test_device.read_fields(['model', 'programmode'], 0)
            except HeatmiserResponseError as err:
                logging.info("C%i device not found, library error %s"%(address, err))
            else:
                model = test_device.model.read_value_text()
                prog_mode = test_device.programmode.read_value_text()
                logging.info("C%i device %s found, with program %s"%(address, model, prog_mode))
                controllersettings = {
                    'address': address,
                    'expected_model': model,
                    'expected_prog_mode': prog_mode
                }
                new_device = self.add_device("C%i"%address, controllersettings, self._setup.settings['devicesgeneral'])
                self.controllers.append(new_device)
                self._addresses_in_use.append(address)
    
    def get_stat_address(self, shortname):
        """Get network address from device name."""
        if isinstance(shortname, basestring):
            return self._statlist[shortname]['address']
        else:
            return shortname

    def set_current_controller_by_name(self, name):
        """Set the current device by name"""
        self._current = getattr(self, name)
        
    def set_current_controller_by_index(self, index):
        """Set the current device by id"""
        self._current = self.controllers[index]
        
    def get_controller_by_name(self, name):
        """Get device if from name"""
        return getattr(self, name)

    def run_method_on_all(self, method, *args, **kwargs):
        """Run a method on all devices"""
        results = []
        for obj in self.controllers:
            results.append(getattr(obj, method)(*args, **kwargs))
        return results



"""Heatmiser Adaptor handles serial connection and basic framing for the Heatmiser Protocol"""

import time
import logging
import serial

from .hm_constants import MAX_FRAME_RESP_LENGTH, MIN_FRAME_READ_RESP_LENGTH, DCB_START, FUNC_WRITE, FUNC_READ, BROADCAST_ADDR, FRAME_WRITE_RESP_LENGTH, FR_CONTENTS, RW_LENGTH_ALL, CRC_LENGTH
import framing
from .exceptions import HeatmiserResponseError, HeatmiserResponseErrorCRC
from .logging_setup import csvlist

def retryer(max_retries=3):
    """Decorates reading from and writing to devices, rerunning the methods on failure"""
    def wraps(func):
        """Part of decorator"""
        def inner(*args, **kwargs):
            """Part of decorator"""
            lasterror = None
            for i in range(max_retries):
                if i is not 0:
                    logging.warn("Gen retrying due to %s"%str(lasterror))
                try:
                    result = func(*args, **kwargs)
                except HeatmiserResponseError as err:
                    lasterror = err
                    continue
                else:
                    return result
            raise HeatmiserResponseError("Failed after %i retries on %s"%(max_retries, str(lasterror)))
        return inner
    return wraps

class HeatmiserAdaptor(object):
    """Handles configuration serial port and provides low level read and write functions"""
    def __init__(self, setup):

        # Initialize setup and get settings
        self._setup = setup
        settings = self._setup.settings

        self.serport = serial.Serial()
        self.serport.bytesize = serial.EIGHTBITS #COM_SIZE
        self.serport.parity = serial.PARITY_NONE #COM_PARITY
        self.serport.stopbits = serial.STOPBITS_ONE #COM_STOP

        self.lastsendtime = None
        self.creationtime = time.time()

        self._update_settings(settings)

        self.lastreceivetime = self.creationtime - self.serport.COM_BUS_RESET_TIME # so that system will get on with sending straight away
        
        if self.auto_connect:
            self.connect()
        
    def __del__(self):
        self._disconnect()

    def _update_settings(self, settings):
        """Check settings and update if needed."""
        for name, value in settings['controller'].iteritems():
            setattr(self, name, value)

        # Configure serial settings after closing if required
        wasopen = False
        if self.serport.isOpen():
            wasopen = True
            self.serport.close() # close port

        for name, value in settings['serial'].iteritems():
            setattr(self.serport, name, value)

        if not self.serport.isOpen() and wasopen:
            self._open_port()

### low level serial commands

    def connect(self):
        """If not open, open serial port and log settings"""
        if not self.serport.isOpen():
            self._open_port()

            logging.info("Gen %s port opened"% (self.serport.portstr))
            logging.debug("Gen %s baud, %s bit, %s parity, with %s stopbits, timeout %s seconds" % (self.serport.baudrate, self.serport.bytesize, self.serport.parity, self.serport.stopbits, self.serport.timeout))
        else:
            logging.warn("Gen serial port was already open")
    
    def _open_port(self):
        """open serial port and logging errors"""
        try:
            self.serport.open()
        except serial.SerialException as err:
            logging.error("Could not open serial port %s: %s" % (self.serport.portstr, err))
            raise
    
    def _disconnect(self):
        """check if serial port is open and if so close"""
        #shouldn't need to called directly because handled by destructor
        if self.serport.isOpen():
            self.serport.close() # close port
            logging.info("Gen serial port closed")
        else:
            logging.warn("Gen serial port was already closed")
    
    def _send_message(self, message):
        """Send message to serial port and log errors"""
        if not self.serport.isOpen():
            self.connect()

        #check time since last received to make sure bus has settled.
        waittime = self.serport.COM_BUS_RESET_TIME - (time.time() - self.lastreceivetime)
        if waittime > 0:
            logging.debug("Gen waiting before sending %.2f"% (waittime))
            time.sleep(waittime)
        
        # http://stackoverflow.com/questions/180606/how-do-i-convert-a-list-of-ascii-values-to-a-string-in-python
        string = ''.join(map(chr, message))

        try:
            self.serport.write(string)    # Write a string
        except serial.SerialTimeoutException as err:
            self.serport.close() #need to close so that isOpen works correctly.
            logging.warning("Write timeout error: %s, sending %s" % (err, csvlist(message)))
            raise
        except serial.SerialException as err:
            self.serport.close() #need to close so that isOpen works correctly.
            logging.warning("Write error: %s, sending %s" % (err, csvlist(message)))
            raise

        self.lastsendtime = time.strftime("%d %b %Y %H:%M:%S +0000", time.localtime(time.time())) #timezone is wrong
        logging.debug("Gen sent %s", csvlist(message))

    def _clear_input_buffer(self):
        """Clears input buffer
        
        Used after CRC check wrong; in case more data was sent than expected."""
    
        time.sleep(self.serport.COM_TIMEOUT) #wait for read timeout to ensure slave finished sending
        try:
            if self.serport.isOpen():
                self.serport.reset_input_buffer() #reset input buffer and dump any contents
            logging.warning("Input buffer cleared")
        except serial.SerialException:
            self.serport.close()
            logging.warning("Failed to clear input buffer")
            raise
    
    def _read_bytes(self, length):
        """read from serial port and log errors"""
        try:
            return self.serport.read(length)
        except serial.SerialException as err:
            #There is no new data from serial port (or port missing) (Doesn't include no response from stat)
            logging.warning("Gen serial port error: %s" % str(err))
            self.serport.close()
            raise
        finally:
            self.serport.timeout = self.serport.COM_TIMEOUT #make sure timeout is reverted
            self.lastreceivetime = time.time() #record last read time. Used to manage bus settling.
    
    def _receive_message(self, length=MAX_FRAME_RESP_LENGTH):
        """Receive message from serial port and log errors
        
        Uses two time outs, one on the first byte and another for full data"""
        if not self.serport.isOpen():
            self.connect()
        logging.debug("Gen listening for %d"%length)
        
        # Listen for the first byte
        timereadstart = time.time()
        self.serport.timeout = self.serport.COM_START_TIMEOUT #wait for start of response
        
        firstbyteread = self._read_bytes(1)

        timereadfirstbyte = time.time()-timereadstart
        logging.debug("Gen waited %.2fs for first byte"%timereadfirstbyte)
        if len(firstbyteread) == 0:
            raise HeatmiserResponseError("No Response")
        
        # Listen for the rest of the response
        self.serport.timeout = max(self.serport.COM_MIN_TIMEOUT, self.serport.COM_TIMEOUT - timereadfirstbyte) #wait for full time out for rest of response, but not less than COM_MIN_TIMEOUT)
        byteread = self._read_bytes(length - 1)

        #Convert back to array
        data = map(ord, firstbyteread) + map(ord, byteread)

        return data

### protocol functions
    
    @retryer(max_retries=3)
    def write_to_device(self, network_address, protocol, unique_address, length, payload):
        """Forms write frame and sends to serial link checking the acknowledgement"""
        #Payload must be list
        msg = framing.form_frame(network_address, protocol, self.my_master_addr,
                                 FUNC_WRITE, unique_address, length, payload)
        try:
            self._send_message(msg)
        except Exception:
            logging.warn("C%i writing to address, no message sent"%(network_address))
            raise

        logging.debug("C%i written to address %i length %i payload %s"%(network_address, unique_address, length, csvlist(payload)))
        if network_address == BROADCAST_ADDR: # if broadcasting force it to wait longer until next send
            self.lastreceivetime = time.time() + self.serport.COM_SEND_MIN_TIME - self.serport.COM_BUS_RESET_TIME
        else: #else listen for acknowledgement
            response = self._receive_message(FRAME_WRITE_RESP_LENGTH)
            try:
                framing.verify_write_ack(protocol, network_address, self.my_master_addr, response)
            except HeatmiserResponseErrorCRC:
                self._clear_input_buffer()
                raise

    def min_time_between_reads(self):
        """Computes the minimum time that adaptor leaves between read commands"""
        return self.serport.COM_BUS_RESET_TIME

    @retryer(max_retries=2)
    def read_from_device(self, network_address, protocol, unique_start_address, expected_length, readall=False):
        """Forms read frame and sends to serial link checking the response"""
        if readall:
            msg = framing.form_read_frame(network_address, protocol, self.my_master_addr, DCB_START, RW_LENGTH_ALL)
            logging.debug("C %i read request to address %i length %i"%(network_address, DCB_START, RW_LENGTH_ALL))
        else:
            msg = framing.form_read_frame(network_address, protocol, self.my_master_addr, unique_start_address, expected_length)
            logging.debug("C %i read request to address %i length %i"%(network_address, unique_start_address, expected_length))
        try: #sending request
            self._send_message(msg)
        except:
            logging.warn("C%i address, read message not sent"%(network_address))
            raise

        time1 = time.time()

        try: #listening for response
            response = self._receive_message(MIN_FRAME_READ_RESP_LENGTH + expected_length)
        except Exception as err:
            logging.warn("C%i read failed from address %i length %i due to %s"%(network_address, unique_start_address, expected_length, str(err)))
            raise

        logging.debug("C%i read in %.2f s from address %i length %i response %s"%(network_address, time.time()-time1, unique_start_address, expected_length, csvlist(response)))

        try: #processing response
            framing.verify_response(protocol, network_address, self.my_master_addr, FUNC_READ, expected_length, response)
        except HeatmiserResponseErrorCRC:
            self._clear_input_buffer()
            raise
        return response[FR_CONTENTS:-CRC_LENGTH]

    def read_all_from_device(self, network_address, protocol, expected_length):
        """Forms read all frame using read_from_device"""
        return self.read_from_device(network_address, protocol, DCB_START, expected_length, True)



"""Generic Heatmiser Device Class

Modules handle all the DCB and self.fields for each device on the Heatmiser network

Ian Horsley 2018
"""
import logging
import time
import copy
import serial

from fields import HeatmiserFieldSingleReadOnly, HeatmiserFieldDoubleReadOnly
from hm_constants import DEFAULT_PROTOCOL, SLAVE_ADDR_MIN, SLAVE_ADDR_MAX
from hm_constants import MAX_AGE_LONG
from hm_constants import FIELD_NAME_LENGTH
from .exceptions import HeatmiserResponseError
from .logging_setup import csvlist

class HeatmiserDevice(object):
    """General device class"""

    ## Initialisation functions and low level functions
    def __init__(self, adaptor, devicesettings, generalsettings=None):
        self._adaptor = adaptor

        # initalise variables
        self.dcb_length = None #set after building fields
        self.floorlimiting = None
        self.lastwritetime = None
        self.lastreadtime = None
        # initalise variables that may be overriden by settings
        self.set_protocol = DEFAULT_PROTOCOL #
        self.set_expected_prog_mode = None
        self.set_long_name = 'Unknown'
        self._load_settings(devicesettings, generalsettings) #take all settings and make them attributes

        # initialise external parameters
        self._buildfields() # add fields to self.fields and insome cases add schdulers (extended regularly)
        self._configure_fields() #build fieldname to number dictionary and attached fields to attributes add dcb address to fields and add set dcb_length  (extended in unknown to change length)
        # estimated read time for read_all method
        self.fullreadtime = self._estimate_read_time(self.dcb_length)
        
        self._set_expected_field_values() #set some fields expected values (extended in week)
        self._connect_observers() #connect various observers methods (extended regularly)
        self.rawdata = [None] * self.dcb_length
    
    def _load_settings(self, settings, generalsettings):
        """Loading settings from dictionary into properties"""
        if generalsettings is not None:
            for name, value in generalsettings.iteritems():
                setattr(self, "set_" + name, value)

        for name, value in settings.iteritems():
            setattr(self, "set_" + name, value)

    def _buildfields(self):
        """build list of fields"""
        self.fields = [
            HeatmiserFieldDoubleReadOnly('DCBlen', 0, [], MAX_AGE_LONG),
            HeatmiserFieldSingleReadOnly('vendor', 2, [0, 1], MAX_AGE_LONG, {'heatmiser': 0, 'OEM': 1}),
            HeatmiserFieldSingleReadOnly('version', 3, [], MAX_AGE_LONG),
            HeatmiserFieldSingleReadOnly('model', 4, [0, 5], MAX_AGE_LONG, {'prt_e_model': 3, 'prt_hw_model': 4, False: 0}),  # DT/DT-E/PRT/PRT-E 00/01/02/03
            HeatmiserFieldSingleReadOnly('address', 11, [SLAVE_ADDR_MIN, SLAVE_ADDR_MAX], MAX_AGE_LONG),
        ]
    
    def _set_expected_field_values(self):
        """set the expected values for fields that should be fixed"""
        self.address.expectedvalue = self.set_address
        self.DCBlen.expectedvalue = self.dcb_length
        self.model.expectedvalue = self.model.readvalues[self.set_expected_model]

    @staticmethod
    def _csvlist_field_names_from(fields):
        """return csv of fieldnames from list of fields"""
        return ', '.join(field.name for field in fields)
        
    def _csvlist_field_names_from_ids(self, fieldids):
        """return csv of fieldnames from list of fieldids"""
        fields = [self.fields[fieldid] for fieldid in fieldids]
        return self._csvlist_field_names_from(fields)
        
    def _configure_fields(self):
        """build dict to map field name to index, map fields tables to properties and set dcb addresses."""
        self.fields.sort(key=lambda field: field.address)
        
        dcbaddress = 0
        self.fieldsbyname = {}
        self._fieldnametonum = {}
        for key, field in enumerate(self.fields):
            #set dcbaddress
            field.dcbaddress = dcbaddress
            dcbaddress += field.fieldlength
            #add field to key lookup
            self._fieldnametonum[field.name] = key
            #store field pointer as property
            setattr(self, field.name, field)
            #store field pointer in dictionary
            self.fieldsbyname[field.name] = field
        #record maximum dcb length
        self.dcb_length = dcbaddress
    
    def _connect_observers(self):
        """called to connect obersers to fields"""
        pass
    
    ## Basic reading and getting functions
    
    def read_raw_data(self, startfieldname, endfieldname):
        """Return subset of raw data"""
        return self.rawdata[getattr(self, startfieldname).dcbaddress:getattr(self, endfieldname).last_dcb_byte_address()]
    
    def read_all(self):
        """Returns all the rawdata having got it from the device"""
        try:
            self.rawdata = self._adaptor.read_all_from_device(self.set_address, self.set_protocol, self.dcb_length)
        except serial.SerialException as err:
            logging.warn("C%i Read all failed, Serial Port error %s"%(self.set_address, str(err)))
            raise

        logging.info("C%i Read all"%(self.set_address))

        self.lastreadtime = time.time()
        self._procpayload(self.rawdata)
        return self.rawdata

    def read_field(self, fieldname, maxage=None):
        """Returns a fields value, gets from the device if to old"""
        return self.read_fields([fieldname], maxage)[0]
    
    def read_fields(self, fieldnames, maxage=None):
        """Returns a list of field values, gets from the device if any are to old"""
        #only get field from network if
        # maxage = None, older than the default from fields
        # maxage = -1, not read before
        # maxage >=0, older than maxage
        # maxage = 0, always
        
        fieldids = [self._fieldnametonum[fieldname] for fieldname in fieldnames if hasattr(self, fieldname) and (maxage == 0 or not getattr(self, fieldname).check_data_fresh(maxage))]
        fieldids = list(set(fieldids)) #remove duplicates, ordering doesn't matter

        if len(fieldids) > 0:
            self._get_fields(fieldids)

        return [self.fieldsbyname[fieldname].get_value() if hasattr(self, fieldname) else None for fieldname in fieldnames]
    
    def get_field_range(self, firstfieldname, lastfieldname=None):
        """gets fieldrange from device
        safe for blocks crossing gaps in dcb"""
        if lastfieldname == None:
            lastfieldname = firstfieldname

        firstfieldid = self._fieldnametonum[firstfieldname]
        lastfieldid = self._fieldnametonum[lastfieldname]
            
        blockstoread = self._get_field_blocks_from_id_range(firstfieldid, lastfieldid)
        fieldstring = firstfieldname.ljust(FIELD_NAME_LENGTH) + " " + lastfieldname.ljust(FIELD_NAME_LENGTH)
        self._get_field_blocks(blockstoread, fieldstring)
    
    def _get_fields(self, fieldids):
        """gets fields from device
        safe for blocks crossing gaps in dcb"""
        blockstoread = self._get_field_blocks_from_id_list(fieldids)
        self._get_field_blocks(blockstoread, self._csvlist_field_names_from_ids(fieldids))
    
    def _get_field_blocks(self, blockstoread, fieldstring):
        """gets field blocks from device
        NOT safe for dcb gaps"""
        #blockstoread list of [field, field, blocklength in bytes]
        estimatedreadtime = self._estimate_blocks_read_time(blockstoread)

        if estimatedreadtime < self.fullreadtime - 0.02: #if to close to full read time, then read all
            try:
                for firstfield, lastfield, blocklength in blockstoread:
                    logging.debug("C%i Reading ui %i to %i len %i, proc %s to %s"%(self.set_address, firstfield.address, lastfield.address, blocklength, firstfield.name, lastfield.name))
                    rawdata = self._adaptor.read_from_device(self.set_address, self.set_protocol, firstfield.address, blocklength)
                    self.lastreadtime = time.time()
                    self._procpartpayload(rawdata, firstfield.name, lastfield.name)
            except serial.SerialException as err:
                logging.warn("C%i Read failed of fields %s, Serial Port error %s"%(self.set_address, fieldstring, str(err)))
                raise
            logging.info("C%i Read fields %s, in %i blocks"%(self.set_address, fieldstring, len(blockstoread)))
        else:
            logging.debug("C%i Read fields %s by read_all, %0.3f %0.3f"%(self.set_address, fieldstring, estimatedreadtime, self.fullreadtime))
            self.read_all()
              
        #data can only be requested from the controller in contiguous blocks
        #functions takes a first and last field and separates out the individual blocks available for the controller type
        #return, fieldstart, fieldend, length of read in bytes
    def _get_field_blocks_from_id_range(self, firstfieldid, lastfieldid):
        """Takes range of fieldids and returns field blocks
        
        Splits by fields by address breaks."""
        blocks = []
        previousfield = None

        for field in self.fields[firstfieldid:lastfieldid + 1]:
            if previousfield is not None and field.address - previousfield.address - previousfield.fieldlength == 0: #if follows previousfield:
                blocks[-1][1] = field
                blocks[-1][2] = field.last_dcb_byte_address() - blocks[-1][0].dcbaddress + 1
            else:
                blocks.append([field, field, field.fieldlength])
            previousfield = field

        return blocks
    
    def _get_field_blocks_from_id_list(self, fieldids):
        """Takes range of fieldids and returns field blocks
        Splits by invalid fields. Uses timing to determine the optimum blocking"""
        #find blocks between lowest and highest field
        fullfieldblock = self._get_field_blocks_from_id_range(min(fieldids), max(fieldids))
        readblocks = []
        for firstfield, lastfield, _ in fullfieldblock:
            #find fields in that block
            inblock = [fieldid for fieldid in fieldids if self._fieldnametonum[firstfield.name] <= fieldid <= self._fieldnametonum[lastfield.name]]
            if len(inblock) > 0:
                #if single read is shorter than individual
                readlen = self.fields[max(inblock)].last_dcb_byte_address() - self.fields[min(inblock)].dcbaddress + 1
                if self._estimate_read_time(readlen) < sum([self._estimate_read_time(self.fields[fieldid].fieldlength) for fieldid in inblock]):
                    readblocks.append([self.fields[min(inblock)], self.fields[max(inblock)], readlen])
                else:
                    for ids in inblock:
                        readblocks.append([self.fields[ids], self.fields[ids], self.fields[ids].fieldlength])
        return readblocks
    
    def _estimate_blocks_read_time(self, blocks):
        """estimates read time for a set of blocks, including the COM_BUS_RESET_TIME between blocks
        excludes the COM_BUS_RESET_TIME before first block"""
        readtimes = [self._estimate_read_time(x[2]) for x in blocks]
        return sum(readtimes) + self._adaptor.min_time_between_reads() * (len(blocks) - 1)
    
    @staticmethod
    def _estimate_read_time(length):
        """"estimates the read time for a call to read_from_device without COM_BUS_RESET_TIME
        based on empirical measurements of one prt_hw_model and 5 prt_e_model"""
        return length * 0.002075 + 0.070727

    def _procfield(self, data, fieldinfo):
        """Process data for a single field storing in relevant."""
        #logging.debug("Processing %s data %s"%(fieldinfo.name, csvlist(data)))
        fieldinfo.update_data(data, self.lastreadtime)

    def _procpartpayload(self, rawdata, firstfieldname, lastfieldname):
        """Wraps procpayload by converting fieldnames to fieldids"""
        #rawdata must be a list
        #converts field names to field numbers to allow process of shortened raw data
        logging.debug("C%i Processing Payload from field %s to %s"%(self.set_address, firstfieldname, lastfieldname))
        firstfieldid = self._fieldnametonum[firstfieldname]
        lastfieldid = self._fieldnametonum[lastfieldname]
        self._procpayload(rawdata, firstfieldid, lastfieldid)
        
    def _procpayload(self, rawdata, firstfieldid=0, lastfieldid=False):
        """Split payload with field information and processes each field"""
        logging.debug("C%i Processing Payload from field %i to %i"%(self.set_address, firstfieldid, lastfieldid))
        if not lastfieldid:
            lastfieldid = len(self.fields)
        
        fullfirstdcbadd = self.fields[firstfieldid].dcbaddress
        
        for field in self.fields[firstfieldid:lastfieldid + 1]:
            length = field.fieldlength
            dcbadd = field.dcbaddress - fullfirstdcbadd #adjust for the start of the request
            
            try:
                self._procfield(rawdata[dcbadd:dcbadd+length], field)
            except HeatmiserResponseError as err:
                logging.warn("C%i Field %s process failed due to %s"%(self.set_address, field.name, str(err)))

        self.rawdata[fullfirstdcbadd:fullfirstdcbadd+len(rawdata)] = rawdata
    
    ## Basic set field functions
    
    def set_field(self, fieldname, values):
        """Set a field (single member of fields) on a device to a state or values. Defined for all known field lengths."""
        #values must not be list for field length 1 or 2
        fieldid = self._fieldnametonum[fieldname]
        field = self.fields[fieldid]
        numericvalues = field.write_value_from_text(values) #convert to numbers if input was text
        
        field.is_writable()
        field.check_values(numericvalues)
        payloadbytes = field.format_data_from_value(numericvalues)
        
        printvalues = numericvalues if isinstance(numericvalues, list) else [numericvalues] #adjust for logging
            
        try:
            self._adaptor.write_to_device(self.set_address, self.set_protocol, field.address, field.fieldlength, payloadbytes)
        except serial.SerialException as err:
            logging.warn("C%i failed to set field %s to %s, due to %s"%(self.set_address, fieldname.ljust(FIELD_NAME_LENGTH), csvlist(printvalues), str(err)))
            raise
        logging.info("C%i set field %s to %s"%(self.set_address, fieldname.ljust(FIELD_NAME_LENGTH), csvlist(printvalues)))
        
        self.lastwritetime = time.time()
        field.update_value(numericvalues, self.lastwritetime)
    
    def set_fields(self, fieldnames, values):
        """Set multiple fields on a device to a state or payload."""
        #It groups adjacent fields and issues multiple sets if required.
        #inputs must be matching length lists
        
        fields = [getattr(self, fieldname) for fieldname in fieldnames if hasattr(self, fieldname)]#Get fields
        outputdata = self._get_payload_blocks_from_list(fields, values)
        try:
            for fields, lengthbytes, payloadbytes, writtenvalues in outputdata:
                logging.debug("C%i Setting ui %i len %i, proc %s to %s"%(self.set_address, fields[0].address, lengthbytes, fields[0].name, fields[-1].name))
                self._adaptor.write_to_device(self.set_address, self.set_protocol, fields[0].address, lengthbytes, payloadbytes)
                self.lastwritetime = time.time()
                self._update_fields_values(writtenvalues, fields)
        except serial.SerialException as err:
            logging.warn("C%i settings failed of fields %s, Serial Port error %s"%(self.set_address, self._csvlist_field_names_from(fields), str(err)))
            raise
        logging.info("C%i set fields %s in %i blocks"%(self.set_address, self._csvlist_field_names_from(fields), len(outputdata)))

    def _update_fields_values(self, values, fields):
        """update the field values once data successfully written"""
        for field, value in zip(fields, values):
            field.update_value(value, self.lastwritetime)
    
    @staticmethod
    def _get_payload_blocks_from_list(fields, values):
        """Converts list of fields and values into groups of payload data"""
        #returns fields, lengthbytes, payloadbytes, values
        sortedfields = sorted(enumerate(fields), key=lambda fielde: fielde[1].address)
        
        valuescopy = copy.deepcopy(values) #force copy of values so doesn't get changed later.
        
        outputdata = []
        previousfield = None
        for orginalindex, field in sortedfields:
            #Check field data and sort values
            field.is_writable()
            field.check_values(valuescopy[orginalindex])
            
            
            if len(outputdata) > 0 and field.dcbaddress - previousfield.last_dcb_byte_address() == 1: #if follows previous field ##Shouldn't this be based on unique address?
                outputdata[-1][0].append(field)
                outputdata[-1][1] += field.fieldlength
                outputdata[-1][2].extend(field.format_data_from_value(valuescopy[orginalindex]))
                outputdata[-1][3].append(valuescopy[orginalindex])

            else:
                outputdata.append([[field], field.fieldlength, field.format_data_from_value(valuescopy[orginalindex]), [valuescopy[orginalindex]]])
            previousfield = field
        return outputdata

DEVICETYPES = {
    None: HeatmiserDevice
}



"""Contains all exceptions used by package"""

class HeatmiserError(RuntimeError):
    """Catch all, not used directly in code."""
    pass

class HeatmiserResponseError(HeatmiserError):
    """Raise this when response from a device is wrong."""
    pass

class HeatmiserResponseErrorCRC(HeatmiserResponseError):
    """Specifically when CRC fails check. This is the most common response error."""
    pass

class HeatmiserControllerSensorError(HeatmiserResponseError):
    """Raise this when controller reports sensor error."""
    pass

class HeatmiserControllerTimeError(HeatmiserError):
    """Raise this when controller time is outside of acceptable limits."""
    pass

class HeatmiserControllerSetupInitError(HeatmiserError):
    """Raise this when init fails."""
    pass



"""generic field definitions for Heatmiser protocol"""
import logging
import time

from hm_constants import BYTEMASK
from .exceptions import HeatmiserResponseError
from .observer import Notifier

VALUES_ON_OFF = {'ON': 1, 'OFF': 0} #assusme that default comes first, need to swtich to ordered dictionary to make it possible to get default value
VALUES_OFF_ON = {'OFF': 0, 'ON': 1}
VALUES_OFF = {'OFF': 0}

class HeatmiserFieldUnknown(Notifier):
    """Class for variable length unknown read only field"""
    writeable = False
    divisor = 1

    def __init__(self, name, address, max_age, length):
        super(HeatmiserFieldUnknown, self).__init__()
        self.name = name
        self.address = address
        self.dcbaddress = address
        self.max_age = max_age
        self.fieldlength = length
        self._reset()

    def __int__(self):
        return self.value

    def __eq__(self, other):
        return self.value == other

    def __repr__(self):
        return str(self.value)

    def __cmp__(self, value):
        return cmp(self.value, value)

    def _reset(self):
        """Reset data and values to unknown."""
        self.data = None
        self.value = None
        self.lastreadtime = None #used to record when the field was last read

    def last_dcb_byte_address(self):
        """returns the address of the last dcb byte"""
        return self.dcbaddress + self.fieldlength - 1

    def update_data(self, data, readtime):
        """update stored data and readtime. Don't compute value because don't know how to map"""
        self.data = data
        self.lastreadtime = readtime

    def update_value(self, value, writetime):
        """Don't update because don't know how to map to data."""
        raise NotImplementedError

    def get_value(self):
        """Return value."""
        return self.value
        
    def is_writable(self):
        """Checks if field is writable"""
        if not self.writeable:
            raise ValueError("set_field: field isn't writeable")

    def check_values(self, values):
        """check a payload matches field spec"""
        raise NotImplementedError

    def check_data_valid(self):
        """check whether data has been set"""
        data_not_valid = self.lastreadtime is None
        if data_not_valid:
            logging.debug("Data item %s not available"%(self.name))
        return not data_not_valid

    def check_data_fresh(self, maxagein=None):
        """check whether data is fresh

        Check field data age is not more than maxage (in seconds)
        maxagein = None, use the default self.maxage
        maxagein = -1, only check if present
        maxagein >=0, use maxagein (0 is effectively always False)
        return False if old, True if recent"""
        if not self.check_data_valid():
            return False
        elif maxagein == -1: #only check present
            return True
        elif maxagein is None: #if none use field defaults
            maxage = self.max_age
        else:
            maxage = maxagein
        #now check time
        if time.time() - self.lastreadtime > maxage:
            logging.debug("Data item %s too old"%(self.name))
            return False
        return True

class HeatmiserField(HeatmiserFieldUnknown):
    """Base class for fields providing basic method calls"""
    #single value and hence single range
    writeable = True
    fieldlength = 0

    def __init__(self, name, address, validrange, max_age, readvalues=None):
        ###valid range list can be [], [min, max], [list of valid values]
        super(HeatmiserField, self).__init__(name, address, max_age, self.fieldlength)
        self.validrange = validrange
        self.value = None
        self.expectedvalue = None
        self.writevalues = self.readvalues = readvalues
        #check isinstance(fieldrange[0], (int, long)) and isinstance(fieldrange[1], (int, long))
        if len(validrange) < 2:
            self.validrange = [0, self.maxdatavalue / self.divisor]

    def is_value(self, name):
        """Returns true if value matches read value."""
        return self.value == self.readvalues[name]

    def is_unknown(self):
        """Returns true if value isn't set."""
        return self.value is None

    def read_value_text(self):
        """returns value converting to label if known"""
        if self.readvalues is None:
            return self.value
        else:
            return self.readvalues.keys()[self.readvalues.values().index(self.value)]

    def write_value_from_text(self, value):
        """maps text to value, otherwise returns input"""
        if self.writevalues is None:
            return value
        else:
            return self.writevalues.get(value, value)

    def update_data(self, data, readtime):
        """update stored data and readtime if data valid. Compute and store value from data."""
        value = self._calculate_value(data)
        if self.expectedvalue is not None and value != self.expectedvalue:
            raise HeatmiserResponseError('Value %i is unexpected for %s, expected %i'%(value, self.name, self.expectedvalue))
        self._validate_range(value)
        self.data = data
        self.value = value
        self.lastreadtime = readtime
        self.notify_value_change(value)

    def update_value(self, value, writetime):
        """Update the field value once successfully written to network"""
        self._validate_range(value, ValueError)
        data = self.format_data_from_value(value)
        self.data = data
        self.value = value
        self.lastreadtime = writetime
        self.notify_value_change(value)

    def _validate_range(self, values, errortype=HeatmiserResponseError, expectedrange=None):
        """validate the value is within range or in list."""
        if expectedrange is None:
            expectedrange = self.validrange

        if len(expectedrange) == 2:
            if values < expectedrange[0] or values > expectedrange[1]:
                raise errortype("Value %.1f  outside expected range (%.1f, %.1f) for %s"%(values, expectedrange[0], expectedrange[1], self.name))
        elif len(expectedrange) > 2:
            if values not in expectedrange:
                raise errortype("Value %.1f  outside expected range %s for %s"%(values, ','.join(map(str, expectedrange)), self.name))
        else:
            raise errortype("Expected range not defined for %s"%(self.name))

    def _calculate_value(self, data):
        """Calculate value from payload bytes"""
        raise NotImplementedError

    def format_data_from_value(self, value):
        """Convert field to byte form for writting to device"""
        raise NotImplementedError

    def check_values(self, values):
        """check a single or double byte field value matches field spec"""
        if not isinstance(values, (int, long)):
            #one or two byte field, not single length values
            raise TypeError("set_field: invalid requested value")

        #checks the values matches the ranges if ranges are defined
        self._validate_range(values, ValueError)

class HeatmiserFieldSingle(HeatmiserField):
    """Class for writable 1 byte field"""
    maxdatavalue = 255
    fieldlength = 1

    def _calculate_value(self, data):
        """Calculate value from payload bytes"""
        return data[0]/self.divisor

    def format_data_from_value(self, value):
        """Convert field to byte form for writting to device"""
        return [value]

class HeatmiserFieldSingleReadOnly(HeatmiserFieldSingle):
    """Class for read only 1 byte field"""
    writeable = False

class HeatmiserFieldDouble(HeatmiserField):
    """Class for writable 2 byte field"""
    maxdatavalue = 65535
    fieldlength = 2

    def _calculate_value(self, data):
        """Calculate value from payload bytes"""
        val_high = data[0]
        val_low = data[1]
        return 1.0*(val_high*256 + val_low)/self.divisor #force float, although always returns integer temps.

    def format_data_from_value(self, value):
        """Convert field to byte form for writting to device"""
        pay_lo = (value & BYTEMASK)
        pay_hi = (value >> 8) & BYTEMASK
        return [pay_lo, pay_hi]

class HeatmiserFieldDoubleReadOnly(HeatmiserFieldDouble):
    """Class for read only 2 byte field"""
    writeable = False

class HeatmiserFieldDoubleReadOnlyTenths(HeatmiserFieldDoubleReadOnly):
    """Class for read only 2 byte field"""
    divisor = 10.0

class HeatmiserFieldMulti(HeatmiserField):
    """Base class for writable multi byte field"""
    maxdatavalue = None

    def _validate_range(self, values, errortype=HeatmiserResponseError, expectedrange=None):
        """validate the value is within range or in list. cyles through list of ranges"""
        for i, item in enumerate(values):
            expectedrange = self.validrange[i % len(self.validrange)]
            super(HeatmiserFieldMulti, self)._validate_range(item, errortype, expectedrange)

    def _calculate_value(self, data):
        """Calculate value from payload bytes"""
        return data

    def format_data_from_value(self, value):
        """Convert field to byte form for writting to device"""
        return list(value) #force a copy

    def check_values(self, values):
        """check a values matches field spec"""
        if len(values) != self.fieldlength:
            #greater than two byte field, values length must match field length
            raise ValueError("set_field: invalid payload length")

        #checks the values matches the ranges if ranges are defined
        self._validate_range(values, ValueError)



"""special field definitions for Heatmiser protocol"""
import logging
import time

from fields import HeatmiserFieldSingle, HeatmiserFieldSingleReadOnly, HeatmiserFieldMulti
from fields import VALUES_ON_OFF
from hm_constants import CURRENT_TIME_DAY, CURRENT_TIME_HOUR, CURRENT_TIME_MIN, CURRENT_TIME_SEC, TIME_ERR_LIMIT
from .exceptions import HeatmiserResponseError, HeatmiserControllerTimeError

class HeatmiserFieldHotWaterVersion(HeatmiserFieldSingleReadOnly):
    """Class for version on hotwater models."""
    floorlimiting = None

    def _calculate_value(self, data):
        """Calculate value from payload bytes"""
        self.floorlimiting = data[0] >> 7
        return data[0] & 0x7f

class HeatmiserFieldHotWaterDemand(HeatmiserFieldSingle):
    """Class to impliment read and write differences for hotwater demand field."""
    def __init__(self, name, address, validrange, max_age):
        super(HeatmiserFieldHotWaterDemand, self).__init__(name, address, validrange, max_age, VALUES_ON_OFF)
        self.writevalues = {'PROG': 0, 'OVER_ON': 1, 'OVER_OFF': 2}

    def update_value(self, value, writetime):
        """Update the field value once successfully written to network if known. Otherwise reset"""
        #handle odd effect on WRITE_hotwaterdemand_PROG

        if value == self.writevalues['PROG']: #returned to program so outcome is unknown
            self._reset()
        elif value == self.writevalues['OVER_OFF']: #if overridden off store the off read value
            super(HeatmiserFieldHotWaterDemand, self).update_value(self.readvalues['OFF'], writetime)
        else:
            super(HeatmiserFieldHotWaterDemand, self).update_value(value, writetime)

class HeatmiserFieldTime(HeatmiserFieldMulti):
    """Class for time field"""
    fieldlength = 4

    def __init__(self, name, address, max_age):
        self.timeerr = None
        validrange = [[1, 7], [0, 23], [0, 59], [0, 59]] #fixed because functions depend on this range.
        super(HeatmiserFieldTime, self).__init__(name, address, validrange, max_age)

    def get_value(self):
        """Return estimated remote time."""
        estimate = time.time() + self.timeerr
        return self.localtimearray(estimate)

    def comparecontrollertime(self):
        """Compare device and local time difference against threshold"""
        # Now do same sanity checking
        # Check the time is within range
        # currentday is numbered 1-7 for M-S
        # localday (python) is numbered 0-6 for Sun-Sat

        if not self.check_data_valid():
            raise HeatmiserResponseError("Time not read before check")

        localtimearray = self.localtimearray(self.lastreadtime) #time that time field was read
        localweeksecs = self._weeksecs(localtimearray)
        remoteweeksecs = self._weeksecs(self.value)
        directdifference = remoteweeksecs - localweeksecs #absolute remote time relative to local time
        wrappeddifferenceup = directdifference - self.DAYSECS * 7 #compute the absolute difference on rollover
        wrappeddifferencedown = directdifference + self.DAYSECS * 7 #compute the absolute difference on rollover
        self.timeerr = min([directdifference, wrappeddifferenceup, wrappeddifferencedown], key=abs)
        logging.debug("Local time %i, remote time %i, error %i"%(localweeksecs, remoteweeksecs, self.timeerr))

        if abs(self.timeerr) > self.DAYSECS:
            raise HeatmiserControllerTimeError("Incorrect day : local is %s, sensor is %s" % (localtimearray[CURRENT_TIME_DAY], self.value[CURRENT_TIME_DAY]))

        if abs(self.timeerr) > TIME_ERR_LIMIT:
            raise HeatmiserControllerTimeError("Time Error %d greater than %d: local is %s, sensor is %s" % (self.timeerr, TIME_ERR_LIMIT, localweeksecs, remoteweeksecs))

    @staticmethod
    def localtimearray(timenow=time.time()):
        """creates an array in heatmiser format for local time. Day 1-7, 1=Monday"""
        #input time.time() (not local)
        localtimenow = time.localtime(timenow)
        nowday = localtimenow.tm_wday + 1 #python tm_wday, range [0, 6], Monday is 0
        nowsecs = min(localtimenow.tm_sec, 59) #python tm_sec range[0, 61]

        return [nowday, localtimenow.tm_hour, localtimenow.tm_min, nowsecs]

    DAYSECS = 86400
    HOURSECS = 3600
    MINSECS = 60

    def _weeksecs(self, localtimearray):
        """calculates the time from the start of the week in seconds from a heatmiser time array"""
        return (localtimearray[CURRENT_TIME_DAY] - 1) * self.DAYSECS + localtimearray[CURRENT_TIME_HOUR] * self.HOURSECS + localtimearray[CURRENT_TIME_MIN] * self.MINSECS + localtimearray[CURRENT_TIME_SEC]

class HeatmiserFieldHeat(HeatmiserFieldMulti):
    """Class for heating schedule field"""
    fieldlength = 12

class HeatmiserFieldWater(HeatmiserFieldMulti):
    """Class for hotwater schedule field"""
    fieldlength = 16



#!/usr/bin/env python
"""Example of setting configurations on Heatmiser stats
Ian Horsley 2018
"""

import logging

from heatmisercontroller.logging_setup import initialize_logger_full
from heatmisercontroller.network import HeatmiserNetwork

initialize_logger_full('logs', logging.INFO)

HMN = HeatmiserNetwork()

### New observations
# Setting holiday replaces overide and on exit returns to prog (not override)

#HMN.Kit.holdTemp(30,21) #mins, temp
#HMN.All.release_hold_temp()

#HMN.release_temp("Cons")
#HMN.All.set_field('holidayhours', 230)
#HMN.All.release_holiday()

#HMN.Kit.set_field('hotwaterdemand', 'PROG')
#quit()
#HMN.All.set_field('onoff', 'ON')
#HMN.All.set_field('runmode', 'FROST')
#HMN.hmUpdateTime(2)

HMN.B1.set_field('runmode', 'HEAT')
HMN.B2.set_field('runmode', 'HEAT')
HMN.Kit.set_field('runmode', 'HEAT')
HMN.Cons.set_field('runmode', 'FROST')
HMN.Sit.set_field('runmode', 'HEAT')

HMN.All.set_field('frosttemp', 10)
HMN.Cons.set_field('frosttemp', 7)

print(HMN.Cons.read_field('frostprotdisable'))
HMN.Cons.set_field('onoff', 'OFF')

#print HMN.All.read_field('holidayhours', 0)
#exit()

#HMN.B2.set_temp(24)
#HMN.Kit.set_temp(24)
#HMN.B1.set_temp(24)
#HMN.Cons.set_temp(24)
#HMN.All.set_temp(24)

#HMN.Kit.release_temp()
#HMN.B1.release_temp()
#HMN.B2.release_temp()
#HMN.Cons.release_temp()

DOWNINACTIVE = 20
DOWNACTIVE = 19
FROST = 12
SLEEP = 16
DAYFROST = [7, 0, FROST]

SPECIAL = [6, 00, 22, 15, 00, DOWNINACTIVE, 21, 00, SLEEP]
WKDAY_ZEARLY = [5, 00, DOWNACTIVE, 8, 30, FROST, 15, 00, DOWNINACTIVE, 21, 00, SLEEP]
WKDAY_ZLATE = [7, 00, DOWNINACTIVE, 12, 00, FROST, 17, 00, DOWNINACTIVE, 21, 30, SLEEP]
WKDAY_ZTRAINING = [6, 30, DOWNACTIVE, 8, 30, FROST, 16, 30, DOWNINACTIVE, 21, 00, SLEEP]
WKEND_ZEARLY = [5, 00, DOWNINACTIVE, 21, 30, SLEEP]
WKEND_ZLATE = [7, 00, DOWNINACTIVE, 21, 30, SLEEP]
WKEND_ZOFF = [6, 00, DOWNINACTIVE, 21, 00, SLEEP]
WKDAY_ZAWAY = [7, 00, DOWNACTIVE, 8, 30, FROST, 17, 30, DOWNINACTIVE, 21, 00, SLEEP]
WKDAY_ZSTARTNIGHT = WKEND_ZOFF

HMN.Kit.set_heating_schedule('mon', WKEND_ZOFF)
HMN.Kit.set_heating_schedule('tues', WKEND_ZOFF)
HMN.Kit.set_heating_schedule('wed', WKEND_ZOFF)
HMN.Kit.set_heating_schedule('thurs', WKEND_ZOFF)
HMN.Kit.set_heating_schedule('fri', WKEND_ZOFF)
HMN.Kit.set_heating_schedule('sat', WKEND_ZOFF)
HMN.Kit.set_heating_schedule('sun', WKEND_ZOFF)

#HMN.Sit.set_heating_schedule('wday', DAYFROST)
#HMN.Sit.set_heating_schedule('wend', DAYFROST)

HMN.Sit.set_heating_schedule('mon', WKEND_ZOFF)
HMN.Sit.set_heating_schedule('tues', WKEND_ZOFF)
HMN.Sit.set_heating_schedule('wed', WKEND_ZOFF)
HMN.Sit.set_heating_schedule('thurs', WKEND_ZOFF)
HMN.Sit.set_heating_schedule('fri', WKEND_ZOFF)
HMN.Sit.set_heating_schedule('sat', WKEND_ZOFF)
HMN.Sit.set_heating_schedule('sun', WKEND_ZOFF)

EVENINGWATER = [17, 30, 18, 30]
NOWATER = [24, 00, 21, 30]
HMN.Kit.set_water_schedule('all', NOWATER)
HMN.Kit.set_water_schedule('tues', EVENINGWATER)
HMN.Kit.set_water_schedule('fri', EVENINGWATER)

UPSLEEP = 16
UPAWAKE = 19
UPWKDAY_ZEARLY = [5, 0, UPAWAKE, 7, 00, FROST, 15, 00, UPAWAKE, 19, 30, UPSLEEP]
UPWKDAY_ZLATE = [6, 0, UPAWAKE, 8, 00, FROST, 17, 00, UPAWAKE, 20, 30, UPSLEEP]
UPWKDAY = [7, 0, UPAWAKE, 8, 30, FROST, 21, 00, UPAWAKE, 21, 30, UPSLEEP]
UPWKEND_ZEARLY = [5, 0, UPAWAKE, 8, 00, FROST, 15, 00, UPAWAKE, 19, 30, UPSLEEP]
UPWKEND_ZLATE = [7, 0, UPAWAKE, 9, 30, FROST, 17, 00, UPAWAKE, 20, 30, UPSLEEP]
UPWKEND = [7, 0, UPAWAKE, 9, 30, FROST, 17, 00, UPAWAKE, 19, 00, UPSLEEP]
UPWKDAY_OLIN = [6, 0, UPAWAKE, 9, 00, FROST, 18, 00, UPAWAKE, 19, 00, UPSLEEP]
DAYFROST = [7, 0, FROST]

HMN.B1.set_heating_schedule('mon', UPWKDAY_ZLATE)
HMN.B1.set_heating_schedule('tues', UPWKDAY_ZLATE)
HMN.B1.set_heating_schedule('wed', UPWKDAY_ZLATE)
HMN.B1.set_heating_schedule('thurs', UPWKDAY_ZLATE)
HMN.B1.set_heating_schedule('fri', UPWKDAY_ZLATE)
HMN.B1.set_heating_schedule('sat', UPWKDAY_ZLATE)
HMN.B1.set_heating_schedule('sun', UPWKDAY_ZLATE)

HMN.B2.set_heating_schedule('mon', UPWKDAY_OLIN)
HMN.B2.set_heating_schedule('tues', UPWKDAY_OLIN)
HMN.B2.set_heating_schedule('wed', UPWKDAY_OLIN)
HMN.B2.set_heating_schedule('thurs', UPWKDAY_OLIN)
HMN.B2.set_heating_schedule('fri', UPWKDAY_OLIN)
HMN.B2.set_heating_schedule('sat', UPWKDAY_OLIN)
HMN.B2.set_heating_schedule('sun', UPWKDAY_OLIN)

HMN.get_controller_by_name('Cons').set_field('wday_heat', [9, 0, 12, 21, 30, 10, 24, 0, 5, 24, 0, 5])
HMN.get_controller_by_name('Cons').set_field('wend_heat', [9, 0, 12, 21, 30, 10, 24, 0, 5, 24, 0, 5])



"""Heatmiser Device Classes

PRT-E Thermostat classes on the Heatmiser network

Ian Horsley 2018
"""
import logging
import time

from genericdevice import HeatmiserDevice, DEVICETYPES
from fields import HeatmiserFieldSingle, HeatmiserFieldSingleReadOnly, HeatmiserFieldDouble, HeatmiserFieldDoubleReadOnly, HeatmiserFieldDoubleReadOnlyTenths
from fields_special import HeatmiserFieldTime, HeatmiserFieldHeat
from fields import VALUES_ON_OFF, VALUES_OFF_ON, VALUES_OFF
from hm_constants import MAX_AGE_LONG, MAX_AGE_MEDIUM, MAX_AGE_SHORT, MAX_AGE_USHORT
from .exceptions import HeatmiserControllerTimeError, HeatmiserControllerSensorError
from schedule_functions import SchedulerDayHeat, SchedulerWeekHeat
from thermostatstate import Thermostat

class ThermoStatWeek(HeatmiserDevice):
    """Device class for thermostats operating weekly programmode
    Heatmiser prt_e_model."""
    is_hot_water = False #returns True if stat is a model with hotwater control, False otherwise
    
    def __init__(self, adaptor, devicesettings, generalsettings=None):
        self.heat_schedule = None #placeholder for heating schedule object
        self.thermostat = None #placeholder for thermostat object
        super(ThermoStatWeek, self).__init__(adaptor, devicesettings, generalsettings)
        #thermostat specific

    def _buildfields(self):
        """add to list of fields"""
        super(ThermoStatWeek, self)._buildfields()
        # list of fields can be sorted by key
        # dcb addresses could be computed from the completed field list and added to field.
        # all should have the first 4 fields, so put these in generic
        self.fields.extend([
            HeatmiserFieldSingleReadOnly('tempformat', 5, [0, 1], MAX_AGE_LONG),  # 00 C,  01 F
            HeatmiserFieldSingleReadOnly('switchdiff', 6, [1, 3], MAX_AGE_LONG),
            HeatmiserFieldSingleReadOnly('frostprotdisable', 7, [0, 1], MAX_AGE_LONG, VALUES_OFF_ON),  #0=enable frost prot when display off,  (opposite in protocol manual,  but tested and user guide is correct)  (default should be enabled)
            HeatmiserFieldDoubleReadOnly('caloffset', 8, [], MAX_AGE_LONG),
            HeatmiserFieldSingleReadOnly('outputdelay', 10, [0, 15], MAX_AGE_LONG),  # minutes (to prevent rapid switching)
            HeatmiserFieldSingleReadOnly('updwnkeylimit', 12, [0, 10], MAX_AGE_LONG),   #limits use of up and down keys
            HeatmiserFieldSingleReadOnly('sensorsavaliable', 13, [0, 4], MAX_AGE_LONG, {'INT_ONLY': 0, 'EXT_ONLY': 1, 'FLOOR_ONLY': 2, 'INT_FLOOR': 3, 'EXT_FLOOR': 4}),  #00 built in only,  01 remote air only,  02 floor only,  03 built in + floor,  04 remote + floor
            HeatmiserFieldSingleReadOnly('optimstart', 14, [0, 3], MAX_AGE_LONG),  # 0 to 3 hours,  default 0
            HeatmiserFieldSingleReadOnly('rateofchange', 15, [], MAX_AGE_LONG),  #number of minutes per degree to raise the temperature,  default 20. Applies to the Wake and Return comfort levels (1st and 3rd)
            HeatmiserFieldSingleReadOnly('programmode', 16, [0, 1], MAX_AGE_LONG, {'day':1, 'week':0}),  #0=5/2,  1= 7day
            HeatmiserFieldSingle('frosttemp', 17, [7, 17], MAX_AGE_LONG),  #default is 12,  frost protection temperature
            HeatmiserFieldSingle('setroomtemp', 18, [5, 35], MAX_AGE_USHORT),
            HeatmiserFieldSingle('floormaxlimit', 19, [20, 45], MAX_AGE_LONG),
            HeatmiserFieldSingleReadOnly('floormaxlimitenable', 20, [0, 1], MAX_AGE_LONG),  #1 is enable, 0 is disable
            HeatmiserFieldSingle('onoff', 21, [0, 1], MAX_AGE_SHORT, VALUES_ON_OFF),  #1 is on, 0 is off
            HeatmiserFieldSingle('keylock', 22, [0, 1], MAX_AGE_SHORT, VALUES_ON_OFF),  #1 is on, 0 is off
            HeatmiserFieldSingle('runmode', 23, [0, 1], MAX_AGE_SHORT, {'HEAT': 0, 'FROST': 1}),   #0 = heating mode,  1 = frost protection mode
            HeatmiserFieldDouble('holidayhours', 24, [0, 720], MAX_AGE_SHORT, VALUES_OFF),  #range guessed and tested,  setting to 0 cancels hold and puts back to program
            #HeatmiserFieldUnknown('unknown', 26, 1, [], MAX_AGE_LONG, 6),  # gap from 26 to 31
            HeatmiserFieldDouble('tempholdmins', 32, [0, 5760], MAX_AGE_SHORT, VALUES_OFF),  #range guessed and tested,  setting to 0 cancels hold and puts setroomtemp back to program
            HeatmiserFieldDoubleReadOnlyTenths('remoteairtemp', 34, [], MAX_AGE_USHORT),  #ffff if no sensor
            HeatmiserFieldDoubleReadOnlyTenths('floortemp', 36, [], MAX_AGE_USHORT),  #ffff if no sensor
            HeatmiserFieldDoubleReadOnlyTenths('airtemp', 38, [], MAX_AGE_USHORT),  #ffff if no sensor
            HeatmiserFieldSingleReadOnly('errorcode', 40, [0, 224, 225, 226], MAX_AGE_SHORT),  # 0 is no error # errors,  0 built in,  1,  floor,  2 remote
            HeatmiserFieldSingleReadOnly('heatingdemand', 41, [0, 1], MAX_AGE_USHORT),  #0 none,  1 heating currently
            HeatmiserFieldTime('currenttime', 43, MAX_AGE_LONG),  #day (Mon - Sun),  hour,  min,  sec. # local estimate should be good, so update once a day
            #5/2 progamming #if hour = 24 entry not used
            HeatmiserFieldHeat('wday_heat', 47, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),  #hour,  min,  temp  (should minutes be only 0 and 30?)
            HeatmiserFieldHeat('wend_heat', 59, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM)
            ])

        self.heat_schedule = SchedulerWeekHeat()
        self.thermostat = Thermostat('Heating', self)

    def _connect_observers(self):
        """connect obersers to fields"""
        super(ThermoStatWeek, self)._connect_observers()
        self.wday_heat.add_notifable_changed(self.heat_schedule.set_raw_field)
        self.wend_heat.add_notifable_changed(self.heat_schedule.set_raw_field)
        
        #on and off
        self.frostprotdisable.add_notifable_is(self.frostprotdisable.readvalues['ON'], self.thermostat.switch_off)
        self.frostprotdisable.add_notifable_is(self.frostprotdisable.readvalues['OFF'], self.thermostat.switch_off)
        self.onoff.add_notifable_is(self.onoff.readvalues['OFF'], self.thermostat.switch_off)
        self.onoff.add_notifable_is(self.onoff.readvalues['ON'], self.thermostat.switch_swap)
        #change within on
        self.setroomtemp.add_notifable_changed(self.thermostat.switch_swap)
        ###don't need the following if not modelling override holds or program events internally
        ###program change event
        self.tempholdmins.add_notifable_is(self.tempholdmins.readvalues['OFF'], self.thermostat.switch_swap) #check triggered by value force change and data change
        self.tempholdmins.add_notifable_is_not(self.thermostat.switch_swap)
        #between frost and setpoint
        self.runmode.add_notifable_is(self.runmode.readvalues['HEAT'], self.thermostat.switch_swap)
        self.runmode.add_notifable_is(self.runmode.readvalues['FROST'], self.thermostat.switch_swap)
        self.holidayhours.add_notifable_is(self.holidayhours.readvalues['OFF'], self.thermostat.switch_swap)
        self.holidayhours.add_notifable_is_not(self.thermostat.switch_swap)

    def _set_expected_field_values(self):
        """set the expected values for fields that should be fixed"""
        super(ThermoStatWeek, self)._set_expected_field_values()
        self.programmode.expectedvalue = self.programmode.readvalues[self.set_expected_prog_mode]

    def _procfield(self, data, fieldinfo):
        """Process data for a single field storing in relevant."""
        super(ThermoStatWeek, self)._procfield(data, fieldinfo)

        if fieldinfo.name == 'currenttime':
            self._checkcontrollertime()

    def _checkcontrollertime(self):
        """run check of device time against local read time, and try to fix if _autocorrectime"""
        try:
            self.currenttime.comparecontrollertime()
        except HeatmiserControllerTimeError as errstr:
            if self.set_autocorrectime is True:
                logging.warn("C%i %s"%(self.set_address, errstr))
                self.set_time()
            else:
                raise

    def get_variables(self):
        """Gets setroomtemp to hotwaterdemand fields from device"""
        self.get_field_range('setroomtemp', 'hotwaterdemand')

    def get_temps_and_demand(self):
        """Gets remoteairtemp to hotwaterdemand fields from device"""
        self.get_field_range('remoteairtemp', 'hotwaterdemand')

    ## External functions for printing data
    def display_heating_schedule(self):
        """Prints heating schedule to stdout"""
        self.heat_schedule.display()

    def nexttarget(self):
        """get next heat target"""
        return self.heat_schedule.get_next_schedule_item(self.currenttime.localtimearray())

    def print_target(self):
        """Returns text describing current heating state"""
        self.read_temp_state()
        return self.thermostat.get_state_text()
            
    ## External functions for reading data

    def read_temp_state(self):
        """Returns the current temperature control state from off to following program"""
        self.read_fields(['mon_heat', 'tues_heat', 'wed_heat', 'thurs_heat', 'fri_heat', 'wday_heat', 'wend_heat'], -1)
        self.read_fields(['onoff', 'frostprotdisable', 'holidayhours', 'runmode', 'tempholdmins', 'setroomtemp'])

        return self.thermostat.state

    def read_air_sensor_type(self):
        """Reports airsensor type"""
        #1 local, 2 remote
        self.read_field('sensorsavaliable')

        if self.sensorsavaliable.is_value('INT_ONLY') or self.sensorsavaliable.is_value('INT_FLOOR'):
            return 1
        elif self.sensorsavaliable.is_value('EXT_ONLY') or self.sensorsavaliable.is_value('EXT_FLOOR'):
            return 2
        raise ValueError("sensorsavaliable field invalid")

    def read_air_temp(self):
        """Read the air temperature getting data from device if too old"""
        if self.read_air_sensor_type() == 1:
            self.read_fields(['airtemp', 'errorcode'], self.set_max_age_temp)
            if self.errorcode == 224:
                raise HeatmiserControllerSensorError("airtemp sensor error")
            return self.airtemp.value
        else:
            self.read_fields(['remoteairtemp', 'errorcode'], self.set_max_age_temp)
            if self.errorcode == 226:
                raise HeatmiserControllerSensorError("remote airtemp sensor error")
            return self.remoteairtemp.value

    def read_time(self, maxage=0):
        """Readtime, getting from device if required"""
        return self.read_field('currenttime', maxage)

    ## External functions for setting data

    def set_heating_schedule(self, day, schedule):
        """Set heating schedule for a single day"""
        padschedule = self.heat_schedule.pad_schedule(schedule)
        for fieldname in self.heat_schedule.get_entry_names(day):
            self.set_field(fieldname, padschedule)

    def set_time(self):
        """set time on device to match current localtime on server"""
        timenow = time.time() + 0.5 #allow a little time for any delay in setting
        return self.set_field('currenttime', self.currenttime.localtimearray(timenow))

    #overriding

    def set_temp(self, temp):
        """sets the temperature demand overriding the program."""
        #Believe it returns at next prog change.
        if self.read_field('tempholdmins') == 0: #check hold temp not applied
            return self.set_field('setroomtemp', temp)
        else:
            logging.warn("%i address, temp hold applied so won't set temp"%(self.set_address))

    def release_temp(self):
        """release setTemp back to the program, but only if temp isn't held for a time (holdTemp)."""
        if self.read_field('tempholdmins') == 0: #check hold temp not applied
            return self.set_field('tempholdmins', 0)
        else:
            logging.warn("%i address, temp hold applied so won't remove set temp"%(self.set_address))

    def hold_temp(self, minutes, temp):
        """sets the temperature demand overrding the program for a set time."""
        #Believe it then returns to program.
        self.set_field('setroomtemp', temp)
        return self.set_field('tempholdmins', minutes)
        #didn't stay on if did minutes followed by temp.

    def release_hold_temp(self):
        """release setTemp or holdTemp back to the program."""
        return self.set_field('tempholdmins', 0)

    def set_holiday(self, hours):
        """sets holiday up for a defined number of hours."""
        return self.set_field('holidayhours', hours)

    def release_holiday(self):
        """cancels holiday mode"""
        return self.set_field('holidayhours', 0)

class ThermoStatDay(ThermoStatWeek):
    """Device class for thermostats operating daily programmode
    Heatmiser prt_e_model."""

    def _buildfields(self):
        """add to list of fields"""
        super(ThermoStatDay, self)._buildfields()
        self.fields.extend([
            HeatmiserFieldHeat('mon_heat', 103, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),
            HeatmiserFieldHeat('tues_heat', 115, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),
            HeatmiserFieldHeat('wed_heat', 127, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),
            HeatmiserFieldHeat('thurs_heat', 139, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),
            HeatmiserFieldHeat('fri_heat', 151, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),
            HeatmiserFieldHeat('sat_heat', 163, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),
            HeatmiserFieldHeat('sun_heat', 175, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM)
        ])

        self.heat_schedule = SchedulerDayHeat()

    def _connect_observers(self):
        """connect obersers to fields"""
        super(ThermoStatDay, self)._connect_observers()
        fieldnames = ['mon_heat', 'tues_heat', 'wed_heat', 'thurs_heat', 'fri_heat', 'sat_heat', 'sun_heat']
        for fieldname in fieldnames:
            getattr(self, fieldname).add_notifable_changed(self.heat_schedule.set_raw_field)

DEVICETYPES.setdefault('prt_e_model', {'week': ThermoStatWeek, 'day': ThermoStatDay})



"""Heatmiser Device Classes

PRT-HW Thermostat classes on the Heatmiser network

Ian Horsley 2018
"""

from genericdevice import DEVICETYPES
from devices_prt_e import ThermoStatWeek, ThermoStatDay
from fields import HeatmiserFieldSingleReadOnly
from fields_special import HeatmiserFieldWater, HeatmiserFieldHotWaterDemand, HeatmiserFieldHotWaterVersion
from hm_constants import MAX_AGE_LONG, MAX_AGE_MEDIUM, MAX_AGE_USHORT
from schedule_functions import SchedulerDayWater, SchedulerWeekWater, SCH_ENT_TEMP

class ThermoStatHotWaterWeek(ThermoStatWeek):
    """Device class for thermostats with hotwater operating weekly programmode
    Heatmiser prt_hw_model."""
    is_hot_water = True

    def __init__(self, adaptor, devicesettings, generalsettings=None):
        self.water_schedule = None #placeholder for hot water schedule
        super(ThermoStatHotWaterWeek, self).__init__(adaptor, devicesettings, generalsettings)

    def _buildfields(self):
        """add to list of fields"""
        super(ThermoStatHotWaterWeek, self)._buildfields()
        
        #thermostat specific version field
        for key, field in enumerate(self.fields):
            if field.name == 'version':
                self.fields[key] = HeatmiserFieldHotWaterVersion('version', 3, [], MAX_AGE_LONG)

        self.fields.extend([
            HeatmiserFieldHotWaterDemand('hotwaterdemand', 42, [0, 2], MAX_AGE_USHORT),  # read [0=off, 1=on],  write [0=as prog, 1=override on, 2=overide off]
            HeatmiserFieldWater('wday_water', 71, [[0, 24], [0, 59]], MAX_AGE_MEDIUM),  # pairs,  on then off repeated,  hour,  min
            HeatmiserFieldWater('wend_water', 87, [[0, 24], [0, 59]], MAX_AGE_MEDIUM)
            #7day progamming
        ])

        self.water_schedule = SchedulerWeekWater()

    def _configure_fields(self):
        """add dummy field for floor limit"""
        super(ThermoStatHotWaterWeek, self)._configure_fields()
        
        field = HeatmiserFieldSingleReadOnly('floorlimiting', None, [0, 1], None) #should specify mapping for on and off.
        
        #store field pointer as property
        setattr(self, field.name, field)
        #store field pointer in dictionary
        self.fieldsbyname[field.name] = field

    def _connect_observers(self):
        """connect obersers to fields"""
        super(ThermoStatHotWaterWeek, self)._connect_observers()
        self.wday_water.add_notifable_changed(self.water_schedule.set_raw_field)
        self.wend_water.add_notifable_changed(self.water_schedule.set_raw_field)

    def _procfield(self, data, fieldinfo):
        """Process data for a single field storing in relevant."""
        super(ThermoStatHotWaterWeek, self)._procfield(data, fieldinfo)
        
        if fieldinfo.name == 'version':
            super(ThermoStatHotWaterWeek, self)._procfield([self.version.floorlimiting], self.floorlimiting)
        
    def display_water_schedule(self):
        """Prints water schedule to stdout"""
        if self.water_schedule is not None:
            self.water_schedule.display()

    ### UNTESTED # last part about scheduletarget doesn't work
    def read_water_state(self):
        """Returns the current hot water control state from off to following program"""
        #does runmode affect hot water state?
        self.read_fields(['mon_water', 'tues_water', 'wed_water', 'thurs_water', 'fri_water', 'wday_water', 'wend_water'], -1)
        self.read_fields(['onoff', 'holidayhours', 'hotwaterdemand'])
        
        if self.onoff == WRITE_ONOFF_OFF:
            return self.TEMP_STATE_OFF
        elif self.holidayhours != 0:
            return self.TEMP_STATE_HOLIDAY
        else:
            self.read_field('currenttime', MAX_AGE_MEDIUM)
            
            locatimenow = self.currenttime.localtimearray()
            scheduletarget = self.water_schedule.get_current_schedule_item(locatimenow)

            if scheduletarget[SCH_ENT_TEMP] != self.hotwaterdemand:
                return self.TEMP_STATE_OVERRIDDEN
        return self.TEMP_STATE_PROGRAM

    def set_water_schedule(self, day, schedule):
        """Set water schedule for a single day"""
        padschedule = self.water_schedule.pad_schedule(schedule)
        for fieldname in self.water_schedule.get_entry_names(day):
            self.set_field(fieldname, padschedule)

class ThermoStatHotWaterDay(ThermoStatDay, ThermoStatHotWaterWeek):
    """Device class for thermostats with hotwater operating daily programmode
    Heatmiser prt_hw_model."""

    def _buildfields(self):
        """add to list of fields"""
        super(ThermoStatHotWaterDay, self)._buildfields()
        self.fields.extend([
            #7day progamming
            HeatmiserFieldWater('mon_water', 187, [[0, 24], [0, 59]], MAX_AGE_MEDIUM),
            HeatmiserFieldWater('tues_water', 203, [[0, 24], [0, 59]], MAX_AGE_MEDIUM),
            HeatmiserFieldWater('wed_water', 219, [[0, 24], [0, 59]], MAX_AGE_MEDIUM),
            HeatmiserFieldWater('thurs_water', 235, [[0, 24], [0, 59]], MAX_AGE_MEDIUM),
            HeatmiserFieldWater('fri_water', 251, [[0, 24], [0, 59]], MAX_AGE_MEDIUM),
            HeatmiserFieldWater('sat_water', 267, [[0, 24], [0, 59]], MAX_AGE_MEDIUM),
            HeatmiserFieldWater('sun_water', 283, [[0, 24], [0, 59]], MAX_AGE_MEDIUM)
        ])
        self.water_schedule = SchedulerDayWater()

    def _connect_observers(self):
        """connect obersers to fields"""
        super(ThermoStatHotWaterDay, self)._connect_observers()
        fieldnames = ['mon_water', 'tues_water', 'wed_water', 'thurs_water', 'fri_water', 'sat_water', 'sun_water']
        for fieldname in fieldnames:
            getattr(self, fieldname).add_notifable_changed(self.water_schedule.set_raw_field)

DEVICETYPES.setdefault('prt_hw_model', {'week': ThermoStatHotWaterWeek, 'day': ThermoStatHotWaterDay})
